REM
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM    *   Grumpy, the most grouchy http server out there!     *
REM    *   Copyright (C) Mateusz Viste 2008, 2009, 2010        *
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM
REM  Written in FreeBASIC v0.21 by Mateusz Viste <mateusz@viste-family.net>
REM
REM   ----------------------------------------------------------------------
REM    This program is free software: you can redistribute it and/or modify
REM    it under the terms of the GNU General Public License as published by
REM    the Free Software Foundation, either version 3 of the License, or
REM    (at your option) any later version.
REM
REM    This program is distributed in the hope that it will be useful,
REM    but WITHOUT ANY WARRANTY; without even the implied warranty of
REM    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
REM    GNU General Public License for more details.
REM
REM    You should have received a copy of the GNU General Public License
REM    along with this program.  If not, see <http://www.gnu.org/licenses/>.
REM   ----------------------------------------------------------------------
REM

#INCLUDE ONCE "vbcompat.bi"           ' Required for use of FileExists(), FORMAT() and FileDateTime() functions
#INCLUDE ONCE "crt.bi"                ' Needed by GetSrvTime
#INCLUDE ONCE "base64.bi"             ' Allows to use the EncBase64() function
#INCLUDE ONCE "sortstr.bi"            ' Sorting lib
#INCLUDE ONCE "MD5Checksum.bas"       ' MD5 functions: createHash() & createFileHash()
#IFDEF __FB_LINUX__
  #INCLUDE ONCE "crt/linux/unistd.bi" ' Required for GetPID()
' #INCLUDE ONCE "zlib.bi"             ' Required to return gziped content to gzip-aware http clients
#ENDIF

DECLARE SUB ReturnFile(FileToReturn AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return301(RedirectTo AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return302(RedirectTo AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return304(ETag AS STRING)
DECLARE SUB Return400(HeadQuery AS BYTE = 0)
DECLARE SUB Return401(HeadQuery AS BYTE = 0, AuthRealm AS STRING)
DECLARE SUB Return403(HeadQuery AS BYTE = 0, CustomMessage AS STRING = "")
DECLARE SUB Return404(MissingFile AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return500(HeadQuery AS BYTE = 0)
DECLARE SUB Return501(HeadQuery AS BYTE = 0)
DECLARE SUB Return505(HeadQuery AS BYTE = 0)
DECLARE SUB About()
DECLARE SUB LogLine(TextToLog AS STRING = "")
DECLARE SUB SetEnv(EnvString AS STRING)
DECLARE SUB ConnClose()
#INCLUDE ONCE "percencoding.bi"       ' Required for the TranslatePercentEnc function
DECLARE FUNCTION RemoveDoubleSlash(StringToRemove AS STRING) AS STRING
DECLARE FUNCTION RemoveDoubleBackSlash(StringToRemove AS STRING) AS STRING
DECLARE FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
DECLARE FUNCTION GetRFC1123time(SerializedGmtDate AS DOUBLE) AS STRING
DECLARE FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
DECLARE FUNCTION HttpSecurityChecks(Query AS STRING) AS STRING
DECLARE FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE
DECLARE FUNCTION CheckForAuth(PathToCheckFor AS STRING) AS STRING
DECLARE FUNCTION LineInput(SocketToListenTo AS INTEGER = 0) AS STRING
DECLARE FUNCTION GetFileSize(FileToCheck AS STRING) AS STRING
DECLARE FUNCTION GetFileDate(FileToCheck AS STRING) AS STRING
DECLARE FUNCTION GetFileDescription(FileToCheck AS STRING, DescrPath AS STRING) AS STRING
DECLARE FUNCTION HtmlizeText(QueryString AS STRING) AS STRING
DECLARE FUNCTION NowGMT() AS DOUBLE
DECLARE FUNCTION GetGMT(SerializedLocalDate AS DOUBLE) AS DOUBLE
DECLARE FUNCTION SlashToBackslash(BYVAL WorkString AS STRING) AS STRING
DECLARE FUNCTION BackslashToSlash(BYVAL WorkString AS STRING) AS STRING
DECLARE FUNCTION CheckTrimBOM(WorkString AS STRING) AS STRING
DECLARE FUNCTION SpaceToNbsp(SpacedText AS STRING) AS STRING
DECLARE FUNCTION GetParentDirectory(GivenDirectory AS STRING) AS STRING
DECLARE FUNCTION ExtractFQDNfromURL(URLtoProcess AS STRING) AS STRING
DECLARE FUNCTION CheckForbiddenExternalReferers(FileMimeType AS STRING) AS BYTE

CONST pVer AS STRING = "0.21"
CONST pDate AS STRING = "2008 - 2011"
CONST homepage AS STRING = "http://www.viste-family.net/mateusz/software/grumpy/"

CONST CRLF AS STRING = CHR(13,10)

' Initializing buffers:
DIM SHARED AS STRING LineBuff, OperationBuff, HostBuff, HostPortBuff, UserAgentBuff, AcceptBuff,_
              AcceptEncodingBuff, AcceptCharsetBuff, KeepAliveBuff, ConnectionBuff, CookieBuff,_
              RefererBuff, IfModifiedSinceBuff, IfUnmodifiedSinceBuff, AuthorizationBuff, RangeBuff,_
              IfNoneMatchBuff, AcceptLangBuff
DIM SHARED AS STRING GetQuery, GetParam1, GetParam2
' Other variables:
REDIM SHARED AS STRING DescriptionTable(0 TO 16)
DIM SHARED AS BYTE VerboseMode, AllowDirListing, CgiSupport
DIM SHARED AS STRING Token, Value, RootPath, LocalFile, SecurityComment, TempString, VirtualHost, SrvSideParams, DirListCSS, LogFile, BlockedMediaFromOtherReferer, RefererWhitelist
DIM SHARED AS STRING xServerAdmin, xHostedBy, xPoweredBy, xDisclaimer, xCustomHeaders
DIM SHARED AS UINTEGER LogEntries, ContentMD5MaxFileSize
DIM SHARED AS ULONGINT QoSlimit
DIM SHARED EnvVariables(1 TO 30) AS STRING
DIM SHARED StartTime AS DOUBLE
DIM SHARED ShortMonths(1 TO 12) AS STRING*3 => {"Jan", "Feb", "Mar", "Apr",_
                                                "May", "Jun", "Jul", "Aug",_
                                                "Sep", "Oct", "Nov", "Dec"}

DIM HeadFlag AS BYTE
DIM TempDouble AS DOUBLE
DIM TempInteger AS INTEGER

#IFDEF __FB_LINUX__
  CONST ConfigFile = "/etc/grumpy.cfg"
  CONST DirSlash = "/"
  CONST CgiExtension = ".cgi"
#ENDIF
#IFDEF __FB_WIN32__
  DIM SHARED ConfigFile AS STRING
  ConfigFile = EXEPATH + "\grumpy.cfg"
  CONST DirSlash = "\"
  CONST CgiExtension = ".exe"
  RANDOMIZE TIMER
  DIM SHARED AS INTEGER RandomPID
  RandomPID = INT(RND*99999)
  FUNCTION GetPID() AS INTEGER
    RETURN RandomPID
  END FUNCTION
#ENDIF

#INCLUDE ONCE "execcgi.bi"   ' SUB ExecCgi(CgiProgram AS STRING, HeadQuery AS BYTE = 0)
#INCLUDE ONCE "getmime.bi"   ' Provides MIME support, via GetMIME(FileToCheck AS STRING)
#INCLUDE ONCE "listdir.bi"   ' SUB ListDir(DirectoryToList AS STRING, HeadFlag AS BYTE = 0)
#INCLUDE ONCE "langnego.bi"  ' FUNCTION CheckForLangNego(LocalFile AS STRING, AcceptLangBuff AS STRING) AS STRING

StartTime = TIMER

VerboseMode = VAL(ReadCFG(ConfigFile, "Verbose"))
AllowDirListing = VAL(ReadCFG(ConfigFile, "AllowDirListing"))
DirListCSS = ReadCFG(ConfigFile, "DirListCSS")
RootPath = ReadCFG(ConfigFile, "RootDir")
QoSlimit = VAL(ReadCFG(ConfigFile, "QoS"))
CgiSupport = VAL(ReadCFG(ConfigFile, "CgiSupport"))
ContentMD5MaxFileSize = VAL(ReadCFG(ConfigFile, "ContentMD5MaxFileSize"))
LogFile = ReadCFG(ConfigFile, "HttpLogFile")

IF LEN(LogFile) = 0 THEN
  #IFDEF __FB_LINUX__
    LogFile = "/var/log/grumpy.log"
  #ENDIF
  #IFDEF __FB_WIN32__
    LogFile = EXEPATH + "\grumpy.log"
  #ENDIF
END IF

IF RootPath = "" THEN
  RootPath = "/var/www/"
  VerboseMode = 1
  AllowDirListing = 0
  QoSlimit = 0
  CgiSupport = 0
END IF

IF RIGHT(RootPath, 1) <> DirSlash THEN RootPath += DirSlash

IF LCASE(COMMAND(1)) = "-h" OR LCASE(COMMAND(1)) = "--help" THEN About()

OPEN CONS FOR INPUT AS #4
IF VerboseMode > 0 THEN
  LogLine()
  TempDouble = NowGMT()
  TempString = ""
  IF LEN(ENVIRON("REMOTE_ADDR")) > 0 THEN TempString = "REMOTE_ADDR=" + ENVIRON("REMOTE_ADDR")
  IF LEN(ENVIRON("REMOTE_HOST")) > 0 THEN TempString = "REMOTE_HOST=" + ENVIRON("REMOTE_HOST") + " " + TempString
  IF LEN(TempString) > 0 THEN TempString = " / " + TRIM(TempString)
  LogLine("---[ " & FORMAT(DAY(TempDouble), "0#") & " " & ShortMonths(MONTH(TempDouble)) & " " & YEAR(TempDouble) & ", " & FORMAT(HOUR(TempDouble), "0#") & ":" & FORMAT(MINUTE(TempDouble), "0#") & ":" & FORMAT(SECOND(TempDouble), "0#") & " (GMT)" & TempString & " ]---")
END IF


TempDouble = TIMER
DO
  LineBuff = LineInput(4)
  IF VerboseMode = 2 THEN LogLine(LineBuff)
  IF TIMER - TempDouble > 20 THEN
    IF VerboseMode > 0 THEN
      LogLine("The request is sent at a suspiciously slow rate (over 20s). The connection has been killed. It could be a SlowLoris-like attack attempt.")
    END IF
    ConnClose()
  END IF

  REM * Perform sanity checks on the request...
  SecurityComment = HttpSecurityChecks(LineBuff)
  IF SecurityComment <> "" THEN
    IF MID(SecurityComment, 1, 1) = "!" THEN
        IF VerboseMode > 0 THEN
          LogLine("***** SECURITY CHECKS *****")
          LogLine(" The security module has detected a problem with the following request:")
          LogLine(" """ + LineBuff + """")
          LogLine(" The request has been processed anyway. You will find a short explanation below.")
          LogLine(" " + CHR(34) + MID(SecurityComment, 2) + CHR(34))
          LogLine("***************************")
        END IF
      ELSE
        IF VerboseMode > 0 THEN
          LogLine("***** SECURITY CHECKS *****")
          LogLine(" The security module has detected a problem with the following request:")
          LogLine(" """ + LineBuff + """")
          LogLine(" The request has been killed. You will find a short explanation below.")
          LogLine(" " + CHR(34) + SecurityComment + CHR(34))
          LogLine("***************************")
        END IF
        ConnClose()
    END IF
  END IF
  REM * Sanity checks done

  Token = UCASE(LEFT(LineBuff, INSTR(LineBuff, " ") - 1))
  Value = TRIM(MID(LineBuff, INSTR(LineBuff, " ") + 1))
  SELECT CASE Token
    CASE "GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS"
      OperationBuff = LineBuff
      GetQuery = TRIM(MID(OperationBuff, 1, INSTR(OperationBuff, " ") - 1))
      IF GetQuery = "HEAD" THEN HeadFlag = 1 ELSE HeadFlag = 0
      GetParam1 = TRIM(MID(MID(OperationBuff, LEN(GetQuery) + 2), 1, INSTR(MID(OperationBuff, LEN(GetQuery) + 2), " ") - 1))
      GetParam2 = TRIM(MID(OperationBuff, LEN(GetQuery) + 3 + LEN(GetParam1)))
    CASE "HOST:"
      HostBuff = Value
      IF INSTR(HostBuff, ":") > 0 THEN
        HostPortBuff = MID(HostBuff, INSTR(HostBuff, ":") + 1)
        HostBuff = MID(HostBuff, 1, INSTR(HostBuff, ":") - 1)
      END IF
    CASE "USER-AGENT:"
      UserAgentBuff = Value
    CASE "ACCEPT:"
      AcceptBuff = Value
    CASE "ACCEPT-ENCODING:"
      AcceptEncodingBuff = Value
    CASE "ACCEPT-CHARSET:"
      AcceptCharsetBuff = Value
    CASE "ACCEPT-LANGUAGE:"
      AcceptLangBuff = Value
    CASE "RANGE:"
      RangeBuff = Value
    CASE "KEEP-ALIVE:"
      KeepAliveBuff = Value
    CASE "CONNECTION:"
      ConnectionBuff = Value
    CASE "REFERER:"
      RefererBuff = Value
    CASE "COOKIE:"
      CookieBuff = Value
    CASE "IF-MODIFIED-SINCE:"
      IfModifiedSinceBuff = Value
    CASE "IF-UNMODIFIED-SINCE:"
      IfUnmodifiedSinceBuff = Value
    CASE "IF-NONE-MATCH:"
      IfNoneMatchBuff = Value
    CASE "AUTHORIZATION:"
      AuthorizationBuff = Value
  END SELECT
LOOP UNTIL (LEN(LineBuff) = 0) OR ((LEN(GetParam1) > 0) AND (LEN(GetParam2) = 0))

REM * Check for banned User-Agents
LineBuff = ReadCFG(ConfigFile, "BannedUserAgents")
WHILE LEN(LineBuff) > 0
  TempInteger = INSTR(LineBuff, "|")
  TempString = ""
  IF TempInteger > 1 THEN
      TempString = MID(LineBuff, 1, TempInteger - 1)
      LineBuff = MID(LineBuff, TempInteger + 1)
    ELSEIF TempInteger = 1 THEN
      LineBuff = MID(LineBuff, 2)
    ELSEIF TempInteger < 1 THEN
      TempString = LineBuff
      LineBuff = ""
  END IF
  IF TRIM(TempString) = "" THEN TempString = ""
  IF LEN(TempString) > 0 THEN
    IF INSTR(LCASE(UserAgentBuff), LCASE(TempString)) > 0 THEN
      IF VerboseMode > 0 THEN LogLine("Banned user-agent detected (" + TempString + ")")
      SLEEP 2000, 1  ' Make him wait a moment
      Return403(HeadFlag)
      ConnClose()
    END IF
  END IF
WEND

REM * Check for virtual hosts
IF LEN(HostBuff) > 0 THEN
  VirtualHost = ReadCFG(ConfigFile, "vhost:" + TRIM(HostBuff))
  IF LEN(VirtualHost) > 0 THEN RootPath = VirtualHost
  IF RIGHT(RootPath, 1) <> DirSlash THEN RootPath += DirSlash
  IF LEFT(RootPath, 2) = "./" OR LEFT(RootPath, 2) = ".\" THEN RootPath = EXEPATH + "/" + MID(RootPath, 3)
  #IFDEF __FB_WIN32__
    RootPath = SlashToBackslash(RootPath)
    RootPath = RemoveDoubleBackslash(RootPath)
  #ENDIF
  #IFDEF __FB_LINUX__
    RootPath = RemoveDoubleSlash(RootPath)
  #ENDIF
END IF

REM * Check for custom headers
xServerAdmin = ReadCFG(ConfigFile, "x-server-admin")
xPoweredBy = ReadCFG(ConfigFile, "x-powered-by")
xHostedBy = ReadCFG(ConfigFile, "x-hosted-by")
xDisclaimer = ReadCFG(ConfigFile, "x-disclaimer")

xCustomHeaders = ""
IF LEN(xPoweredBy) > 0 THEN xCustomHeaders += "X-Powered-By: " + xPoweredBy + CRLF
IF LEN(xHostedBy) > 0 THEN xCustomHeaders += "X-Hosted-By: " + xHostedBy + CRLF
IF LEN(xServerAdmin) > 0 THEN xCustomHeaders += "X-Server-Admin: " + xServerAdmin + CRLF
IF LEN(xDisclaimer) > 0 THEN xCustomHeaders += "X-Disclaimer: " + xDisclaimer
IF RIGHT(xCustomHeaders, 2) = CRLF THEN xCustomHeaders = MID(xCustomHeaders, 1, LEN(xCustomHeaders) - 2)

REM * Process the request...

TempInteger = INSTR(GetParam1, "?")                 '
IF TempInteger > 0 THEN                             '
    SrvSideParams = MID(GetParam1, TempInteger + 1) ' Check for server-side parameters
    GetParam1 = MID(GetParam1, 1, TempInteger - 1)  '
  ELSE                                              '
    SrvSideParams = ""                              '
END IF                                              '

GetQuery = TranslatePercentEnc(GetQuery)            '
GetParam1 = TranslatePercentEnc(GetParam1)          ' Decode % encoding before further treatement
GetParam2 = TranslatePercentEnc(GetParam2)          '


IF VerboseMode > 0 THEN
  TempString = "Query=" + GetQuery + " Param1=" + GetParam1
'  IF LEN(ENVIRON("REMOTE_ADDR")) > 0 THEN TempString = "Remote_Addr=" + ENVIRON("REMOTE_ADDR") + " " + TempString
'  IF LEN(ENVIRON("REMOTE_HOST")) > 0 THEN TempString = "Remote_Host=" + ENVIRON("REMOTE_HOST") + " " + TempString
  IF LEN(SrvSideParams) > 0 THEN TempString += "?" + SrvSideParams
  IF LEN(GetParam2) > 0 THEN TempString += " Param2=" + GetParam2
  IF LEN(VirtualHost) > 0 THEN TempString += " [vhost: " + HostBuff + "]"
  LogLine(TempString)
END IF

IF UCASE(MID(GetParam1, 1, 7)) = "HTTP://" THEN             '
  GetParam1 = MID(GetParam1, 8)                             ' Support for absolute URI request
  IF INSTR(GetParam1, "/") <= 0 THEN                        ' (eg. GET http://www.domain.com/index.htm HTTP/1.1)
      GetParam1 = ""                                        ' It is required by RFC 2616 for any HTTP/1.1 server
    ELSE                                                    ' Note, that the FQDN part of the request is simply
      GetParam1 = MID(GetParam1, INSTR(GetParam1, "/"))     ' ignored.
  END IF                                                    '
END IF

SELECT CASE GetQuery
  CASE "GET", "HEAD"
    SELECT CASE GetParam2
      CASE "HTTP/1.0", "HTTP/1.1"
        LocalFile = RemoveDoubleSlash(RootPath + "/" + GetParam1)
        IF INSTR(LocalFile, "?") > 0 OR INSTR(LocalFile, "*") > 0 THEN
            Return404(GetParam2, HeadFlag)
          ELSE
            #IFDEF __FB_WIN32__
              LocalFile = SlashToBackslash(LocalFile)
              LocalFile = RemoveDoubleBackSlash(LocalFile)
            #ENDIF
            IF CheckForEvasion(RootPath, LocalFile) = 1 THEN
                Return403(HeadFlag)
              ELSE
                IF LEN(HostBuff) > 0 OR UCASE(GetParam2) = "HTTP/1.0" THEN
                    TempString = ReadCFG(ConfigFile, "RedirectDomain:" + HostBuff)
                    IF LEN(TempString) > 0 THEN
                        IF RIGHT(TempString, 1) = DirSlash THEN TempString = LEFT(TempString, LEN(TempString) - 1)
                        IF NOT UCASE(LEFT(TempString, 7)) = "HTTP://" THEN TempString = "http://" + TempString
                        Return301(TempString + GetParam1, HeadFlag)
                      ELSE
                        IF IsDirectory(LocalFile) = 1 THEN
                            IF RIGHT(LocalFile, 1) = DirSlash THEN
                                TempString = CheckForAuth(LocalFile)
                                IF TempString <> "" THEN
                                    Return401(HeadFlag, TempString)
                                  ELSE
                                    IF FileExists(LocalFile + "index.htm") <> 0 THEN
                                        LocalFile += "index.htm"
                                        TempString = CheckForLangNego(LocalFile, AcceptLangBuff)
                                        IF (LEN(TempString) > 0) THEN   ' There is a better lang for the user - redirect him there
                                            Return302(TempString, HeadFlag)
                                          ELSE   ' if no better lang found, return the file
                                            ReturnFile(LocalFile, HeadFlag)
                                        END IF
                                      ELSEIF FileExists(LocalFile + "index.html") <> 0 THEN
                                        LocalFile += "index.html"
                                        TempString = CheckForLangNego(LocalFile, AcceptLangBuff)
                                        IF (LEN(TempString) > 0) THEN   ' There is a better lang for the user - redirect him there
                                            Return302(TempString, HeadFlag)
                                          ELSE   ' if no better lang found, return the file
                                            ReturnFile(LocalFile, HeadFlag)
                                        END IF
                                      ELSEIF AllowDirListing = 1 THEN
                                        ListDir(LocalFile, HeadFlag)
                                      ELSE
                                        Return403(HeadFlag)
                                    END IF
                                END IF
                              ELSE
                                Return301(GetParam1 + "/", HeadFlag)
                            END IF
                          ELSE ' If the ressource is a file
                            TempString = CheckForAuth(LocalFile)
                            IF TempString <> "" THEN
                                Return401(HeadFlag, TempString)
                              ELSE
                                IF (FileExists(LocalFile) <> 0) AND (RIGHT(LocalFile, 13) <> DirSlash + ".grumpy.auth") THEN
                                    IF ((LCASE(RIGHT(LocalFile, 4)) = CgiExtension) AND (CgiSupport = 2)) OR ((MID(LocalFile, 1, LEN(RootPath) + 8) = RootPath + "cgi-bin" + DirSlash) AND (LCASE(RIGHT(LocalFile, 4)) = CgiExtension) AND (CgiSupport > 0)) THEN
                                        #IFDEF __FB_WIN32__       '
                                          Return501(HeadFlag)     ' Not supported on Windows
                                        #ENDIF                    '
                                        #IFDEF __FB_LINUX__
                                          ExecCgi(LocalFile, HeadFlag)
                                        #ENDIF
                                      ELSE
                                        TempString = CheckForLangNego(LocalFile, AcceptLangBuff)
                                        IF (LEN(TempString) > 0) THEN   ' There is a better lang for the user - redirect him there
                                            Return302(TempString, HeadFlag)
                                          ELSE   ' if no better lang found, return the file
                                            ReturnFile(LocalFile, HeadFlag)
                                        END IF
                                    END IF
                                  ELSE
                                    Return404(GetParam1, HeadFlag)
                                END IF
                            END IF
                        END IF
                    END IF
                  ELSE
                    Return400(HeadFlag)
                END IF
            END IF
        END IF
      CASE ELSE
        Return505(HeadFlag)
    END SELECT
  CASE "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS", "CONNECT"
    Return501()
  CASE ELSE
    Return400()
END SELECT

ConnClose()


REM ***  END OF THE MAIN PROGRAM  ***


SUB ReturnFile(FileToReturn AS STRING, HeadFlag AS BYTE = 0)
  DIM AS INTEGER x, QoSdelay, SleepTime, DataFileHandler, RangeNotSatisfiable = 0
  DIM AS ULONGINT FileLength, FirstByte, LastByte, QoScounter
  DIM AS STRING*1 ByteBuff
  DIM AS STRING*1024 KByteBuff
  DIM AS DOUBLE QoStimer, TempDouble
  DIM AS STRING TmpVar, TimeStamp, ETag, ContentCoding, ContentMD5
  DIM AS STRING LastModifiedHeader, ContentType

  TempDouble = FileDateTime(FileToReturn)
  TimeStamp = "" & YEAR(TempDouble) & "-" & MONTH(TempDouble) & "-" & DAY(TempDouble) & "-" & HOUR(TempDouble) & "-" & MINUTE(TempDouble) & "-" & SECOND(TempDouble)
  LastModifiedHeader = GetRFC1123time(GetGMT(TempDouble))
  ContentType = GetMIME(FileToReturn)
  IF CheckForbiddenExternalReferers(ContentType) = 1 THEN
    IF VerboseMode > 0 THEN LogLine("External referers aren't authorised for " + ContentType + " files.")
    Return403(HeadFlag, "external referers prohibited")
    ConnClose()
  END IF

  DataFileHandler = FREEFILE
  OPEN FileToReturn FOR BINARY AS #DataFileHandler
  ETag = CHR(34) & TimeStamp & "-" & LOF(DataFileHandler) & CHR(34)

  IF IfNoneMatchBuff = ETag THEN
      Return304(ETag)
    ELSE
      IF LEN(RangeBuff) > 0 THEN                                                      ' Check validity of the
        IF LCASE(TRIM(MID(RangeBuff, 1, INSTR(RangeBuff, "=") - 1))) <> "bytes" THEN  ' range token's unit
          IF VerboseMode > 0 THEN LogLine("Invalid range unit discarded (" & TRIM(MID(RangeBuff, 1, INSTR(RangeBuff, "=") - 1)) & ")")
          RangeBuff = ""
        END IF
      END IF

      IF LEN(RangeBuff) > 0 THEN
        TmpVar = MID(RangeBuff, INSTR(RangeBuff, "=") + 1)
        FirstByte = VAL(MID(TmpVar, 1, INSTR(TmpVar, "-") - 1)) + 1
        TmpVar = MID(TmpVar, INSTR(TmpVar, "-") + 1)
        x = 0
        DO: x += 1
        LOOP UNTIL MID(TmpVar, x, 1) = "/" OR MID(TmpVar, x, 1) = "," OR x >= LEN(TmpVar)
        IF LEN(MID(TmpVar, 1, x)) > 0 THEN
            LastByte = VAL(MID(TmpVar, 1, x)) + 1
          ELSE
            LastByte = LOF(DataFileHandler)
        END IF
        IF LastByte < FirstByte THEN
          IF VerboseMode > 0 THEN LogLine("Discarding invalid range (" + RangeBuff + ")")
          RangeBuff = ""
          FirstByte = 0
          LastByte = 0
        END IF
        IF LastByte > LOF(DataFileHandler) THEN LastByte = LOF(DataFileHandler)
      END IF

      REM   Here I should start a loop to send away ranges

      IF LEN(RangeBuff) = 0 THEN
          IF VerboseMode > 0 THEN
            IF LEN(ContentType) > 0 THEN
                LogLine("Response: 200 OK (" + FileToReturn + " [" + ContentType + "])")
              ELSE
                LogLine("Response: 200 OK (" + FileToReturn + ")")
            END IF
          END IF
          IF ContentMD5MaxFileSize >= LOF(DataFileHandler) THEN
            ContentMD5 = createFileHash(FileToReturn)
            IF VerboseMode > 1 THEN LogLine("Computed MD5: " + ContentMD5)
          END IF
          PRINT GetParam2 + " 200 OK" + CRLF;
        ELSE
          IF FirstByte > LOF(DataFileHandler) THEN
              IF VerboseMode > 0 THEN LogLine("Response: Requested Range Not Satisfiable (" + RangeBuff + ")")
              PRINT GetParam2 + " 416 Requested Range Not Satisfiable" + CRLF;
              RangeNotSatisfiable = 1
            ELSE
              IF VerboseMode > 0 THEN
                IF LEN(ContentType) > 0 THEN
                    LogLine("Response: 206 Partial Content (" + FileToReturn + " [" + ContentType + "][" & FirstByte & "-" & LastByte & "])")
                  ELSE
                    LogLine("Response: 206 Partial Content (" + FileToReturn + " [" & FirstByte & "-" & LastByte & "])")
                END IF
              END IF
              PRINT GetParam2 + " 206 Partial Content" + CRLF;
          END IF
      END IF

      IF LastByte = 0 THEN LastByte = LOF(DataFileHandler)
      IF FirstByte = 0 THEN FirstByte = 1

      ' Uncomment the line below when GZip support will be okay
      'IF LastByte <= 1024 * 1024 AND LEN(RangeBuff) = 0 AND INSTR(LCASE(AcceptEncodingBuff), "gzip") > 0 THEN ContentCoding = "gzip"

      PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
      PRINT "Server: Grumpy/" + pVer + CRLF;
      PRINT "Connection: close" + CRLF;
      PRINT "Allow: GET, HEAD" + CRLF;
      PRINT "Accept-Ranges: bytes" + CRLF;
      IF RangeNotSatisfiable = 1 THEN
          PRINT "Content-Type: text/html; Charset=UTF-8" + CRLF;
          PRINT "Content-Range: bytes */" & LOF(DataFileHandler) & CRLF;
        ELSE
          IF LEN(ContentCoding) = 0 THEN PRINT "Content-Length: " & (LastByte - FirstByte + 1) & CRLF;
          IF LEN(RangeBuff) > 0 THEN PRINT "Content-Range: bytes " & FirstByte - 1 & "-" & LastByte - 1 & "/" & LOF(DataFileHandler) & CRLF;
          IF LEN(ContentCoding) > 0 THEN PRINT "Content-Coding: " + ContentCoding + CRLF;
          IF LEN(ContentType) > 0 THEN PRINT "Content-Type: " + ContentType + CRLF;
          IF LEN(ContentMD5) > 0 THEN PRINT "Content-MD5: " + EncBase64(ContentMD5) + CRLF;
          PRINT "ETag: " + ETag + CRLF;
          PRINT "Last-Modified: " + LastModifiedHeader + CRLF;
      END IF
      IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
      PRINT CRLF;

      IF HeadFlag = 0 THEN   ' Don't return the file content for HEAD queries
        IF RangeNotSatisfiable = 1 THEN
            PRINT "<html><head><title>416 Requested Range Not Satisfiable</title></head><body><h1>416 - Requested Range Not Satisfiable</h1><br>The page cannot be displayed because the request range was not satisfiable.</body></html>"
          ELSE
            x = FirstByte
            QoScounter = 0
            QoSdelay = 0
            QoStimer = TIMER
            IF ContentCoding = "gzip" THEN  ' Return gziped content for files from 128 bytes up to 1 MiB (if the client supports it)


            ' Here will come the gzipping procedure which will output the gziped content of the file.


              ELSE
                WHILE LastByte - x > 1024                    '
                   x += 1024                                 ' Caching the file
                   GET #DataFileHandler, x - 1024, KByteBuff ' into 1K blocks
                   PRINT KByteBuff;                          '
                   IF QoSlimit > 0 THEN        ' Apply QoS rule, if any
                     QoScounter += 1
                    'IF ABS(TIMER - QoStimer) > 1 THEN QoStimer = TIMER: QoScounter = 0: QoSdelay = 0
                     IF QoScounter >= QoSlimit THEN
                       SleepTime = 1000 * (1 - ABS(TIMER - QoSTimer))
                       IF SleepTime > 0 THEN SLEEP SleepTime, 1
                       IF (SleepTime > 100) AND (QoSdelay = 0) THEN QoSdelay = 1
                       SELECT CASE SleepTime
                         CASE IS > 750
                           QoSdelay *= 3
                         CASE IS > 500
                           QoSdelay *= 1.7
                         CASE IS > 250
                           QoSdelay *= 1.1
                         CASE IS > 200
                           QoSdelay += 1
                         CASE IS < 150
                          QoSdelay -= 1
                       END SELECT
                       IF QoSdelay < 0 THEN QoSdelay = 0
                       QoScounter = 0
                       QoStimer = TIMER
                     END IF
                     IF QoSdelay > 0 THEN SLEEP QoSdelay, 1
                   END IF
                WEND
                FOR x = x TO LastByte
                  GET #DataFileHandler, x, ByteBuff
                  PRINT ByteBuff;
                NEXT x
            END IF
        END IF
      END IF
      REM  Here should be stopped the loop which would return partial ranges
  END IF
  CLOSE #DataFileHandler
END SUB


SUB Return301(RedirectTo AS STRING, HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = "<html><head><title>301</title></head><body><h1>301 - Moved Permanently</h1><br>The document has been moved <a href=""" + RedirectTo + """>here</a>.</body></html>" + CRLF
  IF VerboseMode > 0 THEN LogLine("Response: 301 (" + RedirectTo + ")")
  PRINT GetParam2 + " 301" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Location: " + RedirectTo + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return302(RedirectTo AS STRING, HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = "<html><head><title>302</title></head><body><h1>301 - Moved Temporarily</h1><br>The document has been moved <a href=""" + RedirectTo + """>here</a>.</body></html>" + CRLF
  IF VerboseMode > 0 THEN LogLine("Response: 302 (" + RedirectTo + ")")
  PRINT GetParam2 + " 302" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Location: " + RedirectTo + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return304(ETag AS STRING)
  IF VerboseMode > 0 THEN LogLine("Response: 304 Not Modified (ETag: " + Etag + ")")
  PRINT GetParam2 + " 304 Not Modified" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "ETag: " + ETag + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  'PRINT "Content-Length: 8700"; CRLF;
  'PRINT "Content-Type: text/html; charset=UTF-8"; CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
END SUB


SUB Return400(HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = ""
  DIM AS BYTE CustomError = 0
  IF VerboseMode > 0 THEN LogLine("Response: 400 Bad request")
  IF HeadFlag = 0 THEN
    IF FileExists(RootPath + ".errors/400.htm") <> 0 THEN
        CustomError = 1
        OPEN RootPath + ".errors/400.htm" FOR INPUT AS #20
        WHILE NOT EOF(20)
          LINE INPUT #20, LineBuff
          HttpMessage += LineBuff + CRLF
        WEND
        CLOSE #20
      ELSE
        HttpMessage = "<html><head><title>400</title></head><body><h1>400 - Bad request</h1></body></html>" + CRLF
    END IF
  END IF
  PRINT GetParam2 + " 400 Bad request" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  IF CustomError = 0 THEN PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return401(HeadFlag AS BYTE = 0, AuthRealm AS STRING)
  DIM AS STRING HttpMessage = ""
  DIM AS BYTE CustomError = 0
  IF LEN(AuthorizationBuff) > 0 THEN SLEEP 2000, 1  ' Brute-force protection
  IF VerboseMode > 0 THEN LogLine("Response: 401 Unauthorized")
  IF HeadFlag = 0 THEN
    IF FileExists(RootPath + ".errors/401.htm") <> 0 THEN
        CustomError = 1
        OPEN RootPath + ".errors/401.htm" FOR INPUT AS #20
        WHILE NOT EOF(20)
          LINE INPUT #20, LineBuff
          HttpMessage += LineBuff + CRLF
        WEND
        CLOSE #20
      ELSE
        HttpMessage = "<html><head><title>401</title></head><body><h1>401 - Unauthorized</h1></body></html>" + CRLF
    END IF
  END IF
  PRINT GetParam2 + " 401 Unauthorized" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  IF CustomError = 0 THEN PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  PRINT "WWW-Authenticate: Basic realm=" + CHR(34) + AuthRealm + CHR(34) + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return403(HeadFlag AS BYTE = 0, CustomMessage AS STRING = "")
  DIM AS STRING HttpMessage = ""
  DIM AS BYTE CustomError = 0
  IF LEN(CustomMessage) > 0 THEN CustomMessage = " (" + CustomMessage + ")"
  IF VerboseMode > 0 THEN LogLine("Response: 403 Forbidden" + CustomMessage)
  IF HeadFlag = 0 THEN
    IF FileExists(RootPath + ".errors/403.htm") <> 0 THEN
        CustomError = 1
        OPEN RootPath + ".errors/403.htm" FOR INPUT AS #20
        WHILE NOT EOF(20)
          LINE INPUT #20, LineBuff
          HttpMessage += LineBuff + CRLF
        WEND
        CLOSE #20
      ELSE
        HttpMessage = "<html><head><title>403</title></head><body><h1>403 - Forbidden" + CustomMessage + "</h1></body></html>" + CRLF
    END IF
  END IF
  PRINT GetParam2 + " 403 Forbidden" + CustomMessage + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  IF CustomError = 0 THEN PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return404(MissingFile AS STRING, HeadFlag AS BYTE = 0)
  DIM AS STRING LineBuff, HttpMessage = ""
  DIM AS BYTE CustomError = 0
  IF VerboseMode > 0 THEN LogLine("Response: 404 Not found (" + MissingFile + ")")
  IF HeadFlag = 0 THEN
    IF FileExists(RootPath + ".errors/404.htm") <> 0 THEN
        CustomError = 1
        OPEN RootPath + ".errors/404.htm" FOR INPUT AS #20
        WHILE NOT EOF(20)
          LINE INPUT #20, LineBuff
          HttpMessage += LineBuff + CRLF
        WEND
        CLOSE #20
      ELSE
        HttpMessage = "<html><head><title>404</title></head><body><h1>404 - Not found</h1></body></html>" + CRLF
    END IF
  END IF
  PRINT GetParam2 + " 404 Not found" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  IF CustomError = 0 THEN PRINT "Content-Type: text/html;charset=utf-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return500(HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = "<html><head><title>500</title></head><body><h1>500 - Internal server error</h1></body></html>" + CRLF
  IF VerboseMode > 0 THEN LogLine("Response: 500 Internal server error")
  PRINT GetParam2; " 500 Internal server error" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return501(HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = "<html><head><title>501</title></head><body><h1>501 - Not implemented</h1></body></html>" + CRLF
  IF VerboseMode > 0 THEN LogLine("Response: 501 Not implemented")
  PRINT GetParam2 + " 501 Not implemented" + CRLF;
  PRINT "Date: " + GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


SUB Return505(HeadFlag AS BYTE = 0)
  DIM AS STRING HttpMessage = "<html><head><title>505</title></head><body><h1>505 - HTTP version not supported</h1></body></html>" + CRLF
  IF VerboseMode > 0 THEN LogLine("Response: 505 HTTP version not supported")
  PRINT "HTTP/1.1 505 HTTP version not supported" + CRLF;
  PRINT "Date: "; GetRFC1123time(NowGMT()) + CRLF;
  PRINT "Server: Grumpy/" + pVer + CRLF;
  PRINT "Connection: close" + CRLF;
  PRINT "Allow: GET, HEAD" + CRLF;
  PRINT "Accept-Ranges: bytes" + CRLF;
  PRINT "Content-Length: " & LEN(HttpMessage) & CRLF;
  PRINT "Content-Type: text/html; charset=UTF-8" + CRLF;
  IF LEN(xCustomHeaders) > 0 THEN PRINT xCustomHeaders + CRLF;
  PRINT CRLF;
  IF HeadFlag = 0 THEN PRINT HttpMessage;
END SUB


FUNCTION RemoveDoubleSlash(StringToCheck AS STRING) AS STRING
  DIM AS STRING Wynik
  DIM AS INTEGER x
  Wynik = StringToCheck
  WHILE INSTR(Wynik, "//") > 0
    x = INSTR(Wynik, "//")
    Wynik = MID(Wynik, 1, x) + MID(Wynik, x + 2)
  WEND
  RETURN Wynik
END FUNCTION


FUNCTION RemoveDoubleBackSlash(StringToCheck AS STRING) AS STRING
  DIM AS STRING Wynik
  DIM AS INTEGER x
  Wynik = StringToCheck
  WHILE INSTR(Wynik, "\\") > 0
    x = INSTR(Wynik, "\\")
    Wynik = MID(Wynik, 1, x) + MID(Wynik, x + 2)
  WEND
  RETURN Wynik
END FUNCTION


FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
 STATIC CfgTable(1 TO 2, 0 TO 255) AS STRING
 DIM AS STRING CfgReturnString, CfgTmpBuffer
 DIM AS INTEGER Counter

 IF CfgTable(1, 0) <> "init ok" THEN
   DIM AS INTEGER CfgFileHandler, CfgColonPos, Counter
   CfgFileHandler = FREEFILE
   CfgReturnString = ""
   Counter = 0
   CfgTable(1, 0) = "init ok"
   IF DIR(CFGfile) <> "" THEN
      OPEN CFGfile FOR INPUT AS #CfgFileHandler
      DO
         Counter += 1
         LINE INPUT #CfgFileHandler, CfgTmpBuffer
         CfgTmpBuffer = CheckTrimBOM(CfgTmpBuffer)
         IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
           CfgColonPos = INSTR(CfgTmpBuffer, "=")
           CfgTable(1, Counter) = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
           CfgTable(2, Counter) = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
         END IF
      LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
      CLOSE #CfgFileHandler
   END IF
   CfgTable(2, 0) = STR(Counter)
 END IF

 Counter = 0
 DO
   Counter += 1
   IF UCASE(CfgTable(1, Counter)) = UCASE(CFGField) THEN CfgReturnString = CfgTable(2, Counter)
 LOOP UNTIL CfgReturnString <> "" OR Counter >= VAL(CfgTable(2, 0))

 RETURN CfgReturnString
END FUNCTION


FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
  DIM AS BYTE Wynik
  IF INSTR(EvadingFile, "..") > 0 THEN Wynik = 1 ELSE Wynik = 0
  RETURN Wynik
END FUNCTION


FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE
  DIM AS BYTE Wynik
  DIM AS UINTEGER GetAttribs
  DIM AS STRING TempString, RessourceToCheck = ElementToCheck
  IF RIGHT(RessourceToCheck, 1) = DirSlash THEN RessourceToCheck = LEFT(RessourceToCheck, LEN(RessourceToCheck) - 1)
  TempString = DIR(RessourceToCheck, &h37, GetAttribs)
  IF GetAttribs AND &h10 THEN Wynik = 1 ELSE Wynik = 0
  Return Wynik
END FUNCTION


FUNCTION GetRFC1123time(SerializedGmtDate AS DOUBLE) AS STRING
  REM  Eg: "Sun, 06 Nov 1994 08:49:37 GMT"
  DIM ShortWeekday(1 TO 7) AS STRING => {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"}
  DIM Wynik AS STRING
  Wynik = ShortWeekday(WEEKDAY(SerializedGmtDate, 2)) & ", " & FORMAT(DAY(SerializedGmtDate), "0#") & " " & ShortMonths(MONTH(SerializedGmtDate)) & " " & YEAR(SerializedGmtDate) & " " & FORMAT(HOUR(SerializedGmtDate), "0#") & ":" & FORMAT(MINUTE(SerializedGmtDate), "0#") & ":" & FORMAT(SECOND(SerializedGmtDate), "0#") & " GMT"
  RETURN Wynik
END FUNCTION


FUNCTION NowGMT() AS DOUBLE
  DIM AS DOUBLE Wynik
  DIM AS INTEGER res, YearGMT, MonthGMT, DayGMT, HourGMT, MinuteGMT, SecondGMT
  DIM AS STRING timestr
  DIM AS time_t rawtime
  DIM AS tm Ptr timeptr
  time_(@rawtime)
  timeptr=gmtime(@rawtime)
  timestr=Space(128)
'  res=strftime(Strptr(timestr),128,"%a, %d %b %Y %H:%M:%S GMT",timeptr)
  res = strftime(Strptr(timestr),128,"%Y",timeptr)
  YearGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%m",timeptr)
  MonthGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%d",timeptr)
  DayGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%H",timeptr)
  HourGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%M",timeptr)
  MinuteGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%S",timeptr)
  SecondGMT = VAL(LEFT(timestr, res))
  Wynik = DATESERIAL(YearGMT, MonthGMT, DayGMT) + TIMESERIAL(HourGMT, MinuteGMT, SecondGMT)
  RETURN Wynik
END FUNCTION


FUNCTION GetGMT(SerializedLocalDate AS DOUBLE) AS DOUBLE
  DIM AS DOUBLE DiffGMT, Wynik
  DiffGMT = NowGMT() - NOW
  Wynik = SerializedLocalDate + DiffGMT
  RETURN Wynik
END FUNCTION


FUNCTION HttpSecurityChecks(Query AS STRING) AS STRING
  REM  This function performs various sanity checks on the given query.
  REM  It returns an empty string if checks are Ok. Otherwise, a string
  REM  containing a human description of the problem will be returned.
  DIM AS STRING Wynik
  DIM AS INTEGER x, WarnOnly
  IF UCASE(MID(Query, 1, 8)) = "REFERER:" THEN WarnOnly = 1
  IF UCASE(MID(Query, 1, 7)) = "COOKIE:" THEN WarnOnly = 1
  IF INSTR(UCASE(Query), "%2E") > 0 THEN Wynik = "Evasion attempt using twice-encoded '.' character (%2E)."
  IF INSTR(LCASE(Query), "%1u") > 0 THEN Wynik = "Evasion attempt using %1u encoding of the '.' character. Some very old versions of IIS web server understand %1u as '.' character."
  IF INSTR(LCASE(Query), "%u") > 0 THEN Wynik = "The request contains a %u encoding, which is a specific (not standard) Microsoft format used to encode wide characters. This encoding may be used by an attacker to trick systems based on character string recognition."
  IF UCASE(MID(Query, 1, 4)) = "GET " OR UCASE(MID(Query, 1, 5)) = "HEAD " THEN
    IF INSTR(Query, "?") <> INSTRREV(Query, "?") THEN Wynik = "The request contains more than one '?' character. There's no reasonable explanation for such query."
  END IF
  IF INSTR(MID(Query, INSTRREV(Query, " ") - 2, 2), "%") > 0 THEN Wynik = "The request is ended by an invalid %-encoded char."
  IF VAL(TranslatePercentEnc(Query, 9)) > 0 THEN Wynik = "The request contains at least one specific (not supported) %-encoded character. Grumpy doesn't know how to decode it, therefore it could lead to some evasion attempts."
  IF LEN(Query) > 4096 THEN Wynik = "Possible buffer overflow. The request is longer than 4096 bytes. It may be an attempt to insert fraudulent executable code."
  REM Check for non-ASCII characters
  FOR x = 0 TO LEN(Query) - 1
    IF Query[x] > 127 THEN Wynik = "The HTTP request does not contain only 7-bit ASCII characters (the only characters expected in this type of use). Some attackers seek to exploit implementation flaws in certain browsers and web servers by creating specially-crafted HTTP messages to enable executing arbitrary code on the targeted host."
  NEXT x
  REM Check for control characters
  FOR x = 0 TO LEN(Query) - 1
    IF Query[x] < 32 AND Query[x] <> 10 AND Query[x] <> 13 THEN Wynik = "The HTTP request contains some suspect control characters (less than #32 and other than #10 and #13). Some attackers seek to exploit implementation flaws in certain browsers and web servers by creating specially-crafted HTTP messages to enable executing arbitrary code on the targeted host. The detected character is: #" & Query[x] & "."
  NEXT x
  IF INSTR(Query, "../") > 0 THEN Wynik = "Directory traversal attempt. An URL containing two dots ("".."") has been detected."
  IF INSTR(Query, "./") > 0 THEN Wynik = "An URL containing a dot-slash (""./"") has been detected."
  IF INSTR(Query, "\") > 0 THEN Wynik = "A backslash has been detected in the query. There is no reason for that character to appear in a http request."
  IF WarnOnly = 1 AND LEN(Wynik) > 0 THEN Wynik = "!" + Wynik
  RETURN Wynik
END FUNCTION


SUB About()
  PRINT "Grumpy v"; pVer; " Copyright (C) Mateusz Viste "; pDate
  PRINT
  PRINT "This program is free software: you can redistribute it and/or modify it under"
  PRINT "the terms of the GNU General Public License as published by the Free Software"
  PRINT "Foundation, either version 3 of the License, or (at your option) any later"
  PRINT "version."
  PRINT "This program is distributed in the hope that it will be useful, but WITHOUT ANY"
  PRINT "WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A"
  PRINT "PARTICULAR PURPOSE. See the GNU General Public License for more details."
  PRINT
  PRINT "Grumpy is a simple open-source web server for Linux. Grumpy is not a standalone"
  PRINT "daemon - it requires an inetd-compatible superserver to work. Why one would"
  PRINT "prefer Grumpy over any other Linux web server, like Apache, Lighttp, Jigsaw...?"
  PRINT "Well, there is no reason - I wrote Grumpy for fun. Grumpy doesn't support any"
  PRINT "bells and twistles: the only HTTP commands it knows are ""GET"" and ""HEAD"" (one"
  PRINT "could say it makes it very secure). Grumpy is entirely written in FreeBASIC."
  PRINT
  PRINT "homepage: "; homepage
  PRINT
  END
END SUB


FUNCTION CheckForAuth(PathToCheckFor AS STRING) AS STRING
  DIM AS STRING AuthFile, Realm = "", LineBuff, AuthUser
  DIM AS BYTE CharBuff
  DIM AS INTEGER FileHook
  AuthFile = PathToCheckFor
  DO
    DO
      CharBuff = AuthFile[LEN(AuthFile) - 1]
      IF CharBuff <> ASC(DirSlash) THEN AuthFile = MID(AuthFile, 1, LEN(AuthFile) - 1)
    LOOP UNTIL CharBuff = ASC(DirSlash)

    AuthFile += ".grumpy.auth"
    IF FileExists(AuthFile) <> 0 THEN
        IF VerboseMode > 0 THEN LogLine("Auth: The location requires some credentials")
        FileHook = FREEFILE
        OPEN AuthFile FOR INPUT AS FileHook
        IF LOF(FileHook) > 0 THEN
          DO
            LINE INPUT #FileHook, TempString
            TempString = CheckTrimBOM(TempString)
          LOOP UNTIL EOF(FileHook) OR UCASE(MID(TRIM(TempString), 1, 6)) = "REALM="
          IF UCASE(MID(TRIM(TempString), 1, 6)) = "REALM=" THEN Realm = TRIM(MID(TRIM(TempString), 7))
        END IF
        CLOSE FileHook
        IF Realm = "" THEN Realm = "Authorization required"
      ELSE
        AuthFile = MID(AuthFile, 1, INSTRREV(AuthFile, DirSlash) - 1)
    END IF
  LOOP UNTIL LEN(AuthFile) <= LEN(RootPath) OR LEN(Realm) > 0
  IF LEN(Realm) > 0 AND LEN(AuthorizationBuff) > 0 THEN
    AuthUser = MID(AuthorizationBuff, INSTRREV(AuthorizationBuff, " ") + 1)
    FileHook = FREEFILE
    OPEN AuthFile FOR INPUT AS FileHook
    DO
      LINE INPUT #FileHook, LineBuff
      LineBuff = CheckTrimBOM(LineBuff)
      IF LEFT(LCASE(LineBuff), 6) <> "realm=" AND EncBase64(LineBuff) = AuthUser THEN
        Realm = ""
        IF VerboseMode > 0 THEN LogLine("Auth: Access granted (" + AuthUser + ")")
      END IF
    LOOP UNTIL Realm = "" OR EOF(FileHook)
    CLOSE #FileHook
  END IF
  IF Realm <> "" AND VerboseMode > 0 THEN
    IF LEN(AuthUser) > 0 THEN
        LogLine("Auth: Access denied (" + AuthUser + ")")
      ELSE
        LogLine("Auth: Access denied (no credentials provided)")
    END IF
  END IF
  RETURN Realm
END FUNCTION


FUNCTION LineInput(SocketToListenTo AS INTEGER = 0) AS STRING
  DIM AS STRING Wynik, CharBuff
  DIM AS DOUBLE LastData, StartTime
  DIM AS BYTE TimeoutAbort = 0
  LastData = TIMER
  StartTime = TIMER
  DO
    CharBuff = INPUT(1, SocketToListenTo)
    IF CharBuff <> CHR(13) AND CharBuff <> CHR(10) AND CharBuff <> "" THEN Wynik += CharBuff: LastData = TIMER
    IF TIMER - LastData > 10 THEN TimeoutAbort = 1  ' 10 seconds without data is not acceptable.
    IF TIMER - StartTime > 20 THEN TimeoutAbort = 2 ' 20 seconds is the limit to send a line
  LOOP UNTIL TimeoutAbort > 0 OR LEN(Wynik) > 1024 * 1024 OR CharBuff = CHR(10)
  IF LEN(Wynik) > 1024 * 1024 THEN
    IF VerboseMode > 0 THEN LogLine("I/O OVERFLOW - Treatement aborted. Dataline too long (>1 MiB).")
    ConnClose()
  END IF
  IF TimeoutAbort > 0 THEN
    IF VerboseMode > 0 THEN
      SELECT CASE TimeoutAbort
        CASE 1
          LogLine("I/O TIMEOUT - Treatement aborted. Waiting period to long (>10s).")
        CASE 2
          LogLine("Timeout: The sender was sending a single line for over 20s. It could be a Slowloris-like attack attempt.")
      END SELECT
    END IF
    ConnClose()
  END IF
  RETURN Wynik
END FUNCTION


FUNCTION GetFileSize(FileToCheck AS STRING) AS STRING
  DIM AS ULONGINT FileSize    ' ULONGINT is max 16'777'216 TiB
  DIM AS INTEGER FileHook
  DIM AS STRING SizeString
  FileHook = FREEFILE
  OPEN FileToCheck FOR BINARY AS FileHook
  FileSize = LOF(FileHook)    ' LOF() returns a LONGINT value (max 8'388'608 TiB)
  CLOSE #FileHook
  IF FileSize >= 1024^4 THEN        ' TiB
      SizeString = FORMAT(FileSize / 1024^4, "#.0") & " TiB"
    ELSEIF FileSize >= 1024^3 THEN  ' GiB
      SizeString = FORMAT(FileSize / 1024^3, "#.0") & " GiB"
    ELSEIF FileSize >= 1024^2 THEN  ' MiB
      SizeString = FORMAT(FileSize / 1024^2, "#.0") & " MiB"
    ELSEIF FileSize >= 1024 THEN    ' KiB
      SizeString = FORMAT(FileSize / 1024, "#.0") & " KiB"
    ELSE
      SizeString = FileSize & " bytes"
  END IF
  RETURN SizeString
END FUNCTION


FUNCTION GetFileDate(FileToCheck AS STRING) AS STRING
  DIM MonthList(1 TO 12) AS STRING => {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"}
  DIM AS INTEGER fDay, fMonth, fYear, fHour, fMinute
  DIM AS DOUBLE fTimeSerial
  DIM AS STRING Wynik
  fTimeSerial = FileDateTime(FileToCheck)
  fDay = DAY(fTimeSerial)
  fMonth = MONTH(fTimeSerial)
  fYear = YEAR(fTimeSerial)
  fHour = HOUR(fTimeSerial)
  fMinute = MINUTE(fTimeSerial)
  Wynik = "" & fDay & " " & MonthList(fMonth) & " " & fYear & ", " & fHour & ":" & RIGHT("0" & fMinute, 2)
  RETURN Wynik
END FUNCTION


FUNCTION GetFileDescription(FileToCheck AS STRING, DescrPath AS STRING) AS STRING
  DIM AS STRING Wynik, DescrFile, Token1, Token2, LineBuff
  DIM AS INTEGER FileHook, x
  IF DescriptionTable(0) <> DescrPath THEN
    REDIM DescriptionTable(0 TO 16) AS STRING
    DescriptionTable(0) = DescrPath
    DescrFile = DescrPath + "descript.ion"
    IF FileExists(DescrFile) <> 0 THEN
      FileHook = FREEFILE
      OPEN DescrFile FOR INPUT AS FileHook
        x = 0
        WHILE NOT EOF(FileHook)
          LINE INPUT #FileHook, LineBuff
          LineBuff = CheckTrimBOM(LineBuff)
          LineBuff = TRIM(LineBuff)
          IF LEN(LineBuff) > 0 THEN
            IF x = UBOUND(DescriptionTable) THEN REDIM PRESERVE DescriptionTable(0 TO x * 2) AS STRING
            x += 1
            DescriptionTable(x) = LineBuff
          END IF
        WEND
      CLOSE FileHook
    END IF
  END IF

  x = 0
  DO
    x += 1
    LineBuff = DescriptionTable(x)
    Token1 = TRIM(MID(LineBuff, 1, INSTR(LineBuff, CHR(9)) - 1))
    Token2 = TRIM(MID(LineBuff, INSTR(LineBuff, CHR(9)) + 1))
    #IFDEF __FB_LINUX__
      IF Token1 = FileToCheck THEN Wynik = Token2
      #ENDIF
    #IFDEF __FB_WIN32__   ' Windows shouldn't be case-sensitive
      IF UCASE(Token1) = UCASE(FileToCheck) THEN Wynik = Token2
    #ENDIF
  LOOP UNTIL LEN(Wynik) > 0 OR LEN(LineBuff) = 0 OR x = UBOUND(DescriptionTable)
  RETURN Wynik
END FUNCTION


FUNCTION HtmlizeText(QueryString AS STRING) AS STRING
  DIM AS STRING Processing
  DIM AS INTEGER AmpPosition
  Processing = QueryString
  IF INSTR(QueryString, "&") + INSTR(QueryString, CHR(34)) + INSTR(QueryString, "<") + INSTR(QueryString, ">") > 0 THEN
    FOR AmpPosition = LEN(Processing) - 1 TO 0 STEP -1
      SELECT CASE Processing[AmpPosition]
        CASE 38 ' Ampersand "&"
          Processing = MID(Processing, 1, AmpPosition) + "&amp;" + MID(Processing, AmpPosition + 2, LEN(Processing) - AmpPosition)
        CASE 34 ' Quote """
          Processing = MID(Processing, 1, AmpPosition) + "&quot;" + MID(Processing, AmpPosition + 2, LEN(Processing) - AmpPosition)
        CASE 60 ' Less than "<"
          Processing = MID(Processing, 1, AmpPosition) + "&lt;" + MID(Processing, AmpPosition + 2, LEN(Processing) - AmpPosition)
        CASE 62 ' Greater than ">"
          Processing = MID(Processing, 1, AmpPosition) + "&gt;" + MID(Processing, AmpPosition + 2, LEN(Processing) - AmpPosition)
      END SELECT
    NEXT
  END IF
  RETURN Processing
END FUNCTION


SUB LogLine(TextToLog AS STRING = "")
  DIM AS INTEGER FileSocket
  IF VerboseMode > 0 THEN
    FileSocket = FREEFILE
    IF LogEntries <= 100 THEN
        OPEN LogFile FOR APPEND AS FileSocket
          PRINT #FileSocket, "[PID=" & FORMAT(GetPID(), "00000") & "]"; "  "; TextToLog
        CLOSE FileSocket
      ELSE
        OPEN LogFile FOR APPEND AS FileSocket
          PRINT #FileSocket, "[PID=" & FORMAT(GetPID(), "00000") & "]"; "  "; "*** LOG OVERFLOW - PROCESS ABORTED ***"
        CLOSE FileSocket
        CLOSE
        END
    END IF
  END IF
END SUB


SUB SetEnv(EnvString AS STRING)
  DIM AS BYTE x
  DO
    x += 1
  LOOP UNTIL LEN(EnvVariables(x)) = 0 OR x = UBOUND(EnvVariables)
  IF LEN(EnvVariables(x)) = 0 THEN
    EnvVariables(x) = EnvString
    SETENVIRON(EnvVariables(x))
  END IF
END SUB


FUNCTION SlashToBackslash(BYVAL WorkString AS STRING) AS STRING
  DIM TempVar AS INTEGER
  WHILE INSTR(WorkString, "/") > 0
    SELECT CASE INSTR(WorkString, "/")
      CASE 1
        WorkString = "\" + MID(WorkString, 2)
      CASE LEN(WorkString)
        WorkString = LEFT(WorkString, LEN(WorkString) - 1) + "\"
      CASE ELSE
        WorkString = MID(WorkString, 1, INSTR(WorkString, "/") - 1) + "\" + MID(WorkString, INSTR(WorkString, "/") + 1)
    END SELECT
  WEND
  RETURN WorkString
END FUNCTION


FUNCTION BackslashToSlash(BYVAL WorkString AS STRING) AS STRING
  DIM TempVar AS INTEGER
  WHILE INSTR(WorkString, "\") > 0
    SELECT CASE INSTR(WorkString, "\")
      CASE 1
        WorkString = "/" + MID(WorkString, 2)
      CASE LEN(WorkString)
        WorkString = LEFT(WorkString, LEN(WorkString) - 1) + "/"
      CASE ELSE
        WorkString = MID(WorkString, 1, INSTR(WorkString, "\") - 1) + "/" + MID(WorkString, INSTR(WorkString, "\") + 1)
    END SELECT
  WEND
  RETURN WorkString
END FUNCTION


FUNCTION CheckTrimBOM(WorkString AS STRING) AS STRING  ' Remove any BOM UTF8 header from the line
  DIM Result AS STRING = WorkString
  DIM WorkStringLen AS INTEGER = LEN(WorkString)
  IF WorkStringLen >= 3 THEN
    IF MID(WorkString, 1, 3) = CHR(239) + CHR(187) + CHR(191) THEN
      IF WorkStringLen > 3 THEN
          Result = MID(WorkString, 4)
        ELSE
          Result = ""
      END IF
    END IF
  END IF
  RETURN Result
END FUNCTION


SUB ConnClose()
  CLOSE #4
  IF VerboseMode > 0 THEN LogLine("Connection closed, duration: " & FORMAT(TIMER - StartTime, "0.000") & " s.")
  END
END SUB


FUNCTION SpaceToNbsp(SpacedText AS STRING) AS STRING
  DIM AS INTEGER x
  DIM AS STRING Result = ""
  FOR x = 0 TO LEN(SpacedText) - 1
    IF SpacedText[x] = 32 THEN
        Result += "&nbsp;"
      ELSE
        Result += CHR(SpacedText[x])
    END IF
  NEXT x
  RETURN Result
END FUNCTION


FUNCTION GetParentDirectory(GivenDirectory AS STRING) AS STRING
  REM Returns the parent directory of a given path
  DIM AS STRING Result, WorkString
  IF RIGHT(GivenDirectory, 1) = "/" THEN WorkString = LEFT(GivenDirectory, LEN(GivenDirectory) - 1) ELSE WorkString = GivenDirectory
  SELECT CASE INSTRREV(WorkString, "/")
    CASE 0
      Result = GivenDirectory
    CASE ELSE
      Result = MID(WorkString, 1, INSTRREV(WorkString, "/"))
  END SELECT
  RETURN Result
END FUNCTION


FUNCTION ExtractFQDNfromURL(URLtoProcess AS STRING) AS STRING
  DIM AS STRING TempString = URLtoProcess
  IF UCASE(MID(URLtoProcess, 1, 7)) = "HTTP://" THEN
      TempString = MID(URLtoProcess, 8)
    ELSEIF UCASE(MID(URLtoProcess, 1, 8)) = "HTTPS://" THEN
      TempString = MID(URLtoProcess, 9)
  END IF
  IF INSTR(TempString, "/") > 0 THEN
      TempString = MID(TempString, 1, INSTR(TempString, "/") - 1)
  END IF
  RETURN TempString
END FUNCTION


FUNCTION CheckForbiddenExternalReferers(FileMimeType AS STRING) AS BYTE
  REM * Check against bandwith stealing protection, returns 1 for forbidden files, 0 otherwise.
  DIM AS BYTE Result = 0
  BlockedMediaFromOtherReferer = ReadCFG(ConfigFile, "BlockedMediaFromOtherReferer")
  RefererWhitelist = ReadCFG(ConfigFile, "RefererWhitelist")
  IF LEN(BlockedMediaFromOtherReferer) > 0 AND LEN(HostBuff) > 0 AND LEN(RefererBuff) > 0 THEN
    IF NOT LCASE(ExtractFQDNfromURL(RefererBuff)) = LCASE(HostBuff) THEN
      IF INSTR(LCASE(RefererWhitelist), "(" + LCASE(ExtractFQDNfromURL(RefererBuff)) + ")") = 0 THEN
        REM check if the file is in the list and reject or accept.
        IF INSTR(LCASE(BlockedMediaFromOtherReferer), "(" + LCASE(FileMimeType) + ")") > 0 THEN Result = 1
      END IF
    END IF
  END IF
  RETURN Result
END FUNCTION
